/* rico1.c               Thomas R. Nicely          2009.06.06.1023
 *
 * Freeware copyright (c) 2009 Thomas R. Nicely
 * <http://www.trnicely.net>. No warranties expressed or implied.
 * Distributed under the terms of the GNU GPL and GNU FDL licenses;
 * see <http://www.gnu.org/licenses/licenses.html>. Source,
 * binaries, and license terms available upon request.
 *
 * Searches for prime decuples of the form
 * (p, p+2, p+6, p+8, p+18, p+20, p+30, p+32, p+36, p+38),
 * as suggested by Enrico Federighi (6 May 2008).
 *
 * SYNTAX: rico1 [LB] UB
 *
 * The upper bound UB must be less than 2^32 - 2312, about 1.8e19. The
 * default LB is zero.
 *
 * A typical compile command would be
 *
 * gcc rico1.c trn.c -std=gnu99 -lm -lgmp -D__NOMPFR__ -orico1.exe
 *
 * with trn.c and trn.h present; these support files are available
 * at <http://www.trnicely.net/misc/trn.zip>.
 *
 * The GMP library is required (for the primality testing routine),
 * but MPFR libraries are not needed.
 *
 * Output is sent to the screen and to the file rico1.out.
 *
 */

#include "trn.h"

#define  OUTPUT_FILE  "rico1.out"

int iIsDecuple(uint64_t n);
int iIsPrime(uint64_t n, int iStrength);

static const char signature[]=
  "\n __rico1.c__Version 2009.06.06.1023__Freeware copyright (c) 2009"
  "\n Thomas R. Nicely <http://www.trnicely.net>. No warranties expressed"
  "\n or implied. Distributed under the terms of the GNU GPL and GNU FDL"
  "\n licenses; see <http://www.gnu.org/licenses/licenses.html>. Source,"
  "\n binaries, and license terms available upon request.\n";

/**********************************************************************/
int main(int argc, char **argv)
{
char sz[128], sz2[128];
int nSolutions=0, iStrength=0;
uint64_t n, j, LB, UB;
double t0, t;
long double ldLB, ldUB, ldTemp;
FILE *fpOut;

t0=lfSeconds2() - 0.0005001;
if(argc < 2)
  {
  fprintf(stderr, "%s\n SYNTAX: %s [LB] UB\n", signature, argv[0]);
  exit(EXIT_FAILURE);
  }

/* Set termination procedures for Ctrl+C and Ctrl+Break. */

atexit(vAtExit);
signal(SIGINT, exit);
signal(SIGBREAK, exit);
signal(SIGQUIT, exit);
signal(SIGTERM, exit);
signal(SIGKILL, exit);

/* Use long double conversions for user-friendly input. */

ldLB=0;
if(argc > 2)
  {
  ldLB=floorl(___strtold(argv[1], NULL));
  ldUB=floorl(___strtold(argv[2], NULL));
  }
else
  ldUB=floorl(___strtold(argv[1], NULL));
if(ldLB < 0)ldLB=0;
if(ldUB < 0)ldUB=0;
if(ldLB > ldUB)ldUB=ldLB+ldUB;
if(ldLB > ldUB)
  {
  ldTemp=ldLB;
  ldLB=ldUB;
  ldUB=ldTemp;
  }
if(ldUB > UINT64_MAX - 2311)
  {
  __szULL(sz, UINT64_MAX - 2311);
  fprintf(stderr, "\n ERROR: UB may not exceed %s.", sz);
  exit(EXIT_FAILURE);
  }
LB=floorl(ldLB);
UB=floorl(ldUB);

fpOut=fopen(OUTPUT_FILE, "at");
if(!fpOut)
  {
  fprintf(stderr, "\n ERROR: Unable to open output file " OUTPUT_FILE ".\n");
  exit(EXIT_FAILURE);
  }
fprintf(fpOut, "\n");
fclose(fpOut);

/* Detailed analysis reveals that a _necessary_ condition on p is
   p congruent to 821 or 1451 (mod 2310); otherwise, at least one of
   the ten elements of the cluster will be divisible by 2, 3, 5, 7,
   or 11. */

n=2310*(LB/2310) + 821;
j=0; printf("\n");
while(1)
  {
  if(j % 100000 == 0)
    {
    t=lfSeconds2() - t0;
    __szULL(sz, 1000*(n/1000));
    fprintf(stderr, "\r p = %s  (%.3lf sec)", sz, t);
    fflush(stderr);  /* because above print did not end with newline */
    }
  if(iIsDecuple(n))
    {
    nSolutions++;
    __szULL(sz, n);
    t=lfSeconds2() - t0;
    fprintf(stdout, "\r p = %s is a solution.  (%.3lf sec)\n", sz, t);
    fpOut=fopen(OUTPUT_FILE, "at");
    fprintf(fpOut,
      " p = %s is a solution.  (%.3lf sec)\n", sz, t);
    fclose(fpOut);
    __szULL(sz, n+40);
    fprintf(stderr, "\r p = %s  (%.3lf sec)", sz, t);
    fflush(stderr);
    }
  n += 630;
  if(iIsDecuple(n))
    {
    nSolutions++;
    __szULL(sz, n);
    t=lfSeconds2() - t0;
    fprintf(stdout, "\r p = %s is a solution.  (%.3lf sec)\n", sz, t);
    fpOut=fopen(OUTPUT_FILE, "at");
    fprintf(fpOut, " p = %s is a solution.  (%.3lf sec)\n", sz, t);
    fclose(fpOut);
    __szULL(sz, n+40);
    fprintf(stderr, "\r p = %s  (%.3lf sec)", sz, t);
    fflush(stderr);
    }
  n += 1680; j++;
  if(n > UB)
    {
    t=lfSeconds2() - t0;
    __szULL(sz, UB);
    fprintf(stderr, "\r p = %s  (%.3lf sec)", sz, t);
    fflush(stderr);
    break;
    }
  }

t=lfSeconds2() - t0;
__szULL(sz, LB);
__szULL(sz2, UB);
fpOut=fopen(OUTPUT_FILE, "at");
if(nSolutions==0)
  {
  fprintf(stdout,
    "\r ...No solution found in %s <= p <= %s.  (%.3lf sec)",
    sz, sz2, t);
  fprintf(fpOut,
    " ...No solution found in %s <= p <= %s.  (%.3lf sec)\n",
    sz, sz2, t);
  fclose(fpOut);
  }
else
  {
  fprintf(stderr,
    "\r ...The file " OUTPUT_FILE " contains the results.  (%.3lf sec)" , t);
  fprintf(fpOut,
    " ...No other solutions found in %s <= p <= %s.  (%.3lf sec)\n",
    sz, sz2, t);
  }

return(EXIT_SUCCESS);
}
/**********************************************************************/
int iIsDecuple(uint64_t n)
{
/* Check each member of the candidate decuple for primality. As soon as
   one fails, return zero. If all pass, return 1. The validation is
   done in two stages. First, each member is checked with the weak,
   but fast Miller-Rabin test with base two. If all pass, each one
   is checked with a more reliable BPSW based test (no pseudoprimes
   below 1e15, and none currently known). Ideally, the second phase
   should call a deterministic certification test, such as APRCL2 or
   the elliptic curve algorithm, but I don't yet have one coded. This
   strategy seems to yield optimum speed. */

/* Check with Miller-Rabin test, base 2. */

if(!iIsPrime(n    , 0))return(0);
if(!iIsPrime(n + 2, 0))return(0);
if(!iIsPrime(n + 6, 0))return(0);
if(!iIsPrime(n + 8, 0))return(0);
if(!iIsPrime(n +18, 0))return(0);
if(!iIsPrime(n +20, 0))return(0);
if(!iIsPrime(n +30, 0))return(0);
if(!iIsPrime(n +32, 0))return(0);
if(!iIsPrime(n +36, 0))return(0);
if(!iIsPrime(n +38, 0))return(0);

/* Check with a test based on the BPSW algorithm. Someday we may
   have an APRCL2 or elliptic curve test to call here. */

if(!iIsPrime(n    , 1))return(0);
if(!iIsPrime(n + 2, 1))return(0);
if(!iIsPrime(n + 6, 1))return(0);
if(!iIsPrime(n + 8, 1))return(0);
if(!iIsPrime(n +18, 1))return(0);
if(!iIsPrime(n +20, 1))return(0);
if(!iIsPrime(n +30, 1))return(0);
if(!iIsPrime(n +32, 1))return(0);
if(!iIsPrime(n +36, 1))return(0);
if(!iIsPrime(n +38, 1))return(0);

/* At this point, we have a winner. */

return(1);
}
/**********************************************************************/
int iIsPrime(uint64_t n, int iStrength)
{
/* Checks for primality of n. If iStrength=1, a test based on the
   Bailley-PSW test is used; this has not been proven deterministic,
   but no Bailley-PSW pseudoprimes are known, and it has been shown
   that none exists below 10^15. If iStrength=0, the faster (but less
   reliable) Miller-Rabin test with base 2 is used; there are many
   known pseudoprimes for this test, although they still constitute
   a very small fraction of the integers. In any case, a candidate
   decuplet based on these tests should be certified by a deterministic
   test (e.g., the APRCL2 test, or an elliptic curve algorithm). */

int iFirstCall=1;
uint64_t GCD;
mpz_t mpzN;

/* Eliminate integers < 3 and even integers. */

if(n < 2)return(0);
if(n==2)return(1);
if((n & 1)==0)return(0);

/* Check for proper divisors <= 53. */

GCD=ullGCD(n, POP53);  /* Will detect any odd prime <= 53 as a factor */
if((GCD > 1) && (GCD < n))return(0);

/* Apply either the BPSW or SPRP2 test. */

if(iFirstCall)
  {
  mpz_init2(mpzN, 64);  /* Allows for 64-bit integers */
  iFirstCall=0;
  }

if(iStrength)
  return(iIsPrime64(n, 1000));    /* More reliable */
else
  {
  __mpz_set_ull(mpzN, n);
  return(iMillerRabin(mpzN, 2));  /* Faster */
  }
}
/**********************************************************************/
